// Memory.hxx
//
// This a derived class from ObjectPool that provides additional functions
//  for querying objects.  Since it does not contain any additional data,
//  an ObjectPool can simply be cast to a Memory object.
//

#ifndef _Memory_hxx
#define _Memory_hxx

#include <values.h>

#include "common.hxx"
#include "basic.hxx"
#include "Config.hxx"
#include "Address.hxx"
#include "Input.hxx"
#include "Output.hxx"
#include "LongUDPSocket.hxx"
#include "ObjectPool.hxx"
#include "objects.hxx"

namespace Rescue {

class Memory : public ObjectPool
{
public:

  //
  // DATATYPES
  //

  // Object Functions
  struct ObjectFn {
    virtual bool consider(const Object *o) const = 0;
    virtual double f(const Object *o) const = 0;
  };
  
  struct UniformObjectFn : public ObjectFn {
    virtual bool consider(const Object *o) const { return 1; }
    virtual double f(const Object *o) const { return 0.0; }
  };

  struct IsObjectObjectFn : public UniformObjectFn {
    Id m_id;
    virtual bool consider(const Object *o) const { return o->id() == m_id; }
    IsObjectObjectFn(Id id) { m_id = id; }
    IsObjectObjectFn(Object *o) { m_id = o->id(); }
  };

  template <class T>
  struct IsClassObjectFn : public UniformObjectFn {
    virtual bool consider(const Object *o) const { 
      return (dynamic_cast<const T*>(o)); }
  };

  struct IsAgentObjectFn : public Memory::UniformObjectFn {
    virtual bool consider(const Object *o) const { 
      switch(o->type()) { 
      case TYPE_FIRE_BRIGADE: case TYPE_FIRE_STATION:
      case TYPE_AMBULANCE_TEAM: case TYPE_AMBULANCE_CENTER:
      case TYPE_POLICE_FORCE: case TYPE_POLICE_OFFICE:
	return true;
      default:
	return false;
      }
    }
  };

  struct DistanceObjectFn : public ObjectFn {
    double m_x, m_y;
    virtual bool consider(const Object *o) const { return 1; }
    virtual double f(const Object *o) const {
      double x = o->x(); double y = o->y();
      return ((x - m_x) * (x - m_x) + (y - m_y) * (y - m_y));
    }
    DistanceObjectFn(double x, double y) { m_x = x; m_y = y; }
    DistanceObjectFn(const Object *o) { m_x = o->x(); m_y = o->y(); }
  };

  struct AddObjectFn : public ObjectFn {
    const ObjectFn &m_a, &m_b; double m_wa, m_wb;
    virtual bool consider(const Object *o) const {
      return (m_a.consider(o) && m_b.consider(o)); }
    virtual double f(const Object *o) const {
      return (m_a.f(o) * m_wa + m_b.f(o) * m_wb); }
    AddObjectFn(const ObjectFn &a, const ObjectFn &b, 
		double wa = 1.0, double wb = 1.0) : m_a(a), m_b(b) {
      m_wa = wa; m_wb = wb; 
    }
  };

  struct NotObjectFn : public ObjectFn {
    const ObjectFn &m_a;
    virtual bool consider(const Object *o) const { return ! m_a.consider(o); }
    virtual double f(const Object *o) const { return 0.0; }
    NotObjectFn(const ObjectFn &a) : m_a(a) { }
  };

  // Object Pair Functions

  struct ObjectPairFn {
    virtual bool consider(const Object *a, const Object *b) const = 0;
    virtual double f(const Object *a, const Object *b) const = 0;
  };

  struct TraversalObjectPairFn : public ObjectPairFn {
    virtual bool consider(const Object *to, const Object *from) const { 
      const MotionlessObject *f = dynamic_cast<const MotionlessObject*>(from);
      if (!f) return 0;
      if (std::find(f->neighbors().begin(), f->neighbors().end(), to) ==
	  f->neighbors().end()) return 0;

      if (to->type() == TYPE_ROAD) {
	Road *r = (Road *) to;
	int lines = (from == r->head() ? r->linesToTail() : r->linesToHead());
	return (lines - floor((0.5 * (r->linesToHead() + r->linesToTail()) * 
			       (r->block() / (double) r->width())) + 0.5) > 0);
      } else return 1;
    }
    virtual double f(const Object *to, const Object *from) const {
      switch(to->type()) {
      case TYPE_ROAD: return ((Road *) to)->length();
      default: return 0;
      }
    }
  };


  // Other Datatypes

  struct Location {
    Object *m_o;
    S32 m_dist;
    
    Location(Object *o, S32 dist = 0) { m_o = o; m_dist = dist; }
    Location(MovingObject *o) { 
      m_o = o->position(); m_dist = o->positionExtra();
    }
  };
  
  enum FunctionOrder {FUNCTION_MINIMIZE = 0, FUNCTION_MAXIMIZE = 1};
  
  //
  // METHODS
  //

  // Finding Objects

  Object *find(const ObjectFn &q, FunctionOrder order = FUNCTION_MAXIMIZE);

  Object **find_list(const ObjectFn &q, double threshold = -MAXDOUBLE,
		     FunctionOrder order = FUNCTION_MAXIMIZE);

  // Paths

  Object **find_path(const Location &origin,
		     const ObjectFn &destination,
		     const ObjectFn &heuristic =UniformObjectFn(),
		     const ObjectPairFn &edge_cost =TraversalObjectPairFn());
  
  bool verify_path(Object **path, 
		   const ObjectPairFn &edge_cost =TraversalObjectPairFn());

  Object **update_path(Object **path, Object *here,
		       const ObjectFn &destination,
		       const ObjectPairFn &edge_cost =TraversalObjectPairFn());

  Object *path_destination(Object **path);
  Object **path_subpath(Object **path, Object *origin);
  double path_cost(Object **path, 
		   const ObjectPairFn &edge_cost =TraversalObjectPairFn());

private:
  template <class T>
  T *get_typed(Id id) { return dynamic_cast<T *>(get(id)); }
};

inline Object *Memory::path_destination(Object **path) 
{ 
  for( ; *path; path++); return *(path-1); 
}

inline Object **Memory::path_subpath(Object **path, Object *origin)
{
  for( ; *path && *path != origin; path++); return (*path) ? path : NULL; 
}

inline double Memory::path_cost(Object **path, const ObjectPairFn &edge_cost)
{
  double cost = 0.0;
  if (!path[0]) return 0.0;
  for(int i=1; path[i]; i++) cost += edge_cost.f(path[i], path[i-1]);
  return cost;

}

}

#endif
